import pytest
import trio
import mock
import os

from distkv.mock.mqtt import stdtest
from distkv.util import P, Path
from distkv.server import Server
from asyncactor.actor import Actor

import logging

logger = logging.getLogger(__name__)

N = 10
NN = 10
NX = 10

rs = os.environ.get("PYTHONHASHSEED", None)
if rs is None:
    import random
else:
    import trio._core._run as tcr

    random = tcr._r

F1 = 0.05
F2 = 0.05
F3 = 0.05
F4 = 0.2

_asm = Actor._send_msg


async def send_msg(self, msg):
    if random.random() < F1:
        logger.info("NoMSG %s %r", self._name, msg)
        return
    logger.debug("MSG %s %r", self._name, msg)
    await _asm(self, msg)


_aqm = Actor.queue_msg


async def queue_msg(self, msg):
    if random.random() < F2:
        logger.info("NoQUM %s %r", self._name, msg)
        return
    # logger.debug("QUM %s %r",self._name, msg)
    await _aqm(self, msg)


_old_send = Server._send_event


async def send_evt(self, action: str, msg: dict):
    if random.random() < F3:
        logger.info("NoOUT %s %s %r", self.node.name, action, msg)
        return
    logger.debug("OUT %s %s %r", self.node.name, action, msg)
    return await _old_send(self, action, msg)


_old_upm = Server._unpack_multiple


def unpack_multiple(self, msg: dict):
    if random.random() < F4:
        logger.info("NoIN  %s %r", self.node.name, msg)
        return
    logger.debug("IN  %s %r", self.node.name, msg)
    return _old_upm(self, msg)


@pytest.mark.trio
async def test_10_recover(autojump_clock):  # pylint: disable=unused-argument
    """
    This test starts multiple servers at the same time and checks that
    dropping random messages ultimately recovers.
    """
    async with stdtest(test_1={"init": 420}, n=N, tocks=15000) as st:
        st.ex.enter_context(mock.patch("asyncactor.actor.Actor._send_msg", new=send_msg))
        st.ex.enter_context(mock.patch("asyncactor.actor.Actor.queue_msg", new=queue_msg))
        st.ex.enter_context(mock.patch("distkv.server.Server._send_event", new=send_evt))
        st.ex.enter_context(
            mock.patch("distkv.server.Server._unpack_multiple", new=unpack_multiple)
        )

        for x in range(NX):
            for i in range(N):
                async with st.client((i + x) % N) as ci:
                    for j in range(NN):
                        await ci.set(Path("test", i, j), value=(i, j, x))

        await trio.sleep(1)

        missed = True
        while missed:
            missed = False
            for s in range(N):
                async with st.client(s) as ci:
                    c = 0
                    async for r in ci.get_tree(P("test"), min_depth=2, nchain=5):
                        if r.value == (r.path[-2], r.path[-1], NX - 1):
                            c += 1
                        else:
                            logger.info("%d: old %r", s, r)
                    if c != N * NN:
                        res = await ci._request("get_state", iter=False, missing=True)
                        logger.info("%d: missing %d %r", s, N * NN - c, res.missing)
                        missed = True
            if missed:
                await trio.sleep(100)



"""
This module contains the default values for distkv configuration.
"""

from .util import attrdict, combine_dict, NotGiven, P
from .ext import list_ext, load_ext

__all__ = ["PORT", "CFG"]

PORT = 27586  # 20000 + 100*ord('K') + ord('V')

# This default configuration will be used to supplement whatever
# configuration you use.
# It is "complete" in the sense that DistKV will never die
# due to a KeyError caused by a missing config value.

CFG = attrdict(
    logging={  # a magic incantation
        "version": 1,
        "loggers": {"asyncserf": {"level": "INFO"}, "xknx.raw_socket": {"level": "INFO"}},
        "root": {"handlers": ["stderr"], "level": "INFO"},
        "handlers": {
            #           "logfile": {
            #               "class": "logging.FileHandler",
            #               "filename": "test.log",
            #               "level": "DEBUG",
            #               "formatter": "std",
            #           },
            "stderr": {
                "class": "logging.StreamHandler",
                "level": "DEBUG",
                "formatter": "std",
                "stream": "ext://sys.stderr",
            }
        },
        "formatters": {
            "std": {
                "class": "distkv.util.TimeOnlyFormatter",
                "format": "%(asctime)s %(levelname)s:%(name)s:%(message)s",
            }
        },
        "disable_existing_loggers": False,
    },
    connect=attrdict(
        # client: controls how to talk to the DistKV server
        host="localhost",
        port=PORT,
        ssl=False,
        # ssl=attrdict(cert='/path/to/cert.pem',key='/path/to/cert.key'),
        init_timeout=5,  # time to wait for connection plus greeting
        auth=None,  # no auth used by default
        name=None,  # defaults to the server's name
    ),
    config=attrdict(prefix=P(":.distkv.config")),
    errors=attrdict(prefix=P(":.distkv.error")),
    codes=attrdict(prefix=P(":.distkv.code.proc")),
    modules=attrdict(prefix=P(":.distkv.code.module")),
    runner=attrdict(  # for distkv.runner.RunnerRoot
        prefix=P(":.distkv.run"),  # storage for runnable commands
        state=P(":.distkv.state"),  # storage for runner states
        name="run",  # Serf event name, suffixed by subpath
        start_delay=1,  # time to wait between job starts. Not optional.
        ping=-15,  # set an I-am-running message every those-many seconds
        # positive: set in distkv, negative: broadcast to :distkv:run tag
        actor=attrdict(cycle=10, nodes=-1, splits=5),  # Actor config  # required for Runner
        sub=attrdict(group="any", single="at", all="all"),  # tags for various runner modes
    ),
    server=attrdict(
        # server-side configuration
        backend=NotGiven,  # must be specified
        serf=attrdict(
            # how to connect to Serf
            host="localhost",
            port=7373,
        ),
        root=(":distkv",),  # event message name prefix. Should start with a colon.
        paranoia=False,  # typecheck server-to-server updates?
        # which addresses/ports to accept DistKV connections on
        bind=[attrdict()],
        bind_default=attrdict(  # default values for all elements of "bind"
            host="localhost", port=PORT, ssl=False
        ),
        change=attrdict(length=5),  # chain length: use max nr of network sections +1
        ping=attrdict(cycle=10, gap=2),  # asyncserf.Actor config timing for ping
        delete=attrdict(cycle=10, gap=2),  # asyncserf.Actor config timing for deletion
        # ping time also controls minimum startup time
    ),
    paranoia=False,  # typecheck server>server updates?
    # how does a new server reach existing nodes, to download state?
    domain=None,  # domain in which to look up node names, if not in hostmap
    hostmap={  # map DistKV server names to connect destinations
        "test1": ("localhost", PORT),
        "test2": ("does-not-exist.invalid", PORT),
    },
)

for n, _ in list_ext("config"):  # pragma: no cover
    CFG[n] = combine_dict(load_ext(n, "config", "CFG"), CFG.get(n, {}), cls=attrdict)



"""
An online-updated config store

"""

try:
    from contextlib import asynccontextmanager
except ImportError:  # pragma: no cover
    from async_generator import asynccontextmanager

from .errors import ServerError
from .obj import ClientRoot, ClientEntry

import logging

logger = logging.getLogger(__name__)


class ConfigEntry(ClientEntry):
    @classmethod
    def child_type(cls, name):  # pragma: no cover
        """Given a node, return the type which the child with that name should have.
        The default is "same as this class".
        """
        logger.warning("Online config sub-entries are ignored")
        return ClientEntry

    async def set_value(self, value):
        await self.root.client.config._update(self._name, value)


class ConfigRoot(ClientRoot):
    CFG = "config"

    @classmethod
    def child_type(cls, name):
        """Given a node, return the type which the child with that name should have.
        The default is "same as this class".
        """
        return ConfigEntry

    @asynccontextmanager
    async def run(self):
        try:
            async with super().run() as x:
                yield x
        except ServerError:  # pragma: no cover
            logger.exception("No config data")



import anyio
import asyncserf
from contextlib import asynccontextmanager
from . import Backend

# Simply setting connect=asyncserf.serf_client interferes with mocking
# when testing.


class SerfBackend(Backend):
    client = None

    @asynccontextmanager
    async def connect(self, *a, **k):
        async with asyncserf.serf_client(*a, **k) as c:
            self.client = c
            try:
                yield self
            finally:
                self.client = None

    def monitor(self, *topic):  # pylint: disable=invalid-overridden-method
        topic = "user:" + ".".join(topic)
        # self.client.stream is also async, pass thru
        return self.client.stream(topic)

    def send(self, *topic, payload):  # pylint: disable=invalid-overridden-method
        """
        Send this payload to this topic.
        """
        # self.client.event is also async, pass thru
        return self.client.event(".".join(topic), payload=payload, coalesce=False)


@asynccontextmanager
async def connect(*a, **kw):
    async with anyio.create_task_group() as tg:
        c = SerfBackend(tg)
        async with c.connect(*a, **kw):
            yield c



import pytest
import jsonschema
from functools import partial

from distkv.mock import run
from distkv.mock.mqtt import stdtest

from distkv.client import ServerError
from distkv.exceptions import ClientAuthRequiredError, ClientAuthMethodError
from distkv.auth import gen_auth
from distkv.util import P

import logging

logger = logging.getLogger(__name__)


@pytest.mark.trio
async def test_22_auth_basic(autojump_clock):  # pylint: disable=unused-argument
    async with stdtest(args={"init": 123}, tocks=50) as st:
        (s,) = st.s
        h = p = None
        for h, p, *_ in s.ports:
            if h[0] != ":":
                break
        run_c = partial(run, "-D", "client", "-h", h, "-p", p)

        async with st.client() as c:
            assert (await c.get(P(":"))).value == 123

        r = await run_c("data", "get", ":")
        assert r.stdout == "123\n"

        r = await run_c("auth", "-m", "root", "user", "add")

        r = await run_c("data", "get", ":")
        assert r.stdout == "123\n"

        r = await run_c("auth", "-m", "root", "init")

        with pytest.raises(ClientAuthRequiredError):
            await run_c("data", "get", ":")
        with pytest.raises(ClientAuthRequiredError):
            async with st.client() as c:
                assert (await c.get(P(":"))).value == 123

        r = await run_c("-a", "root", "data", "get", ":")
        assert r.stdout == "123\n"

        anull = gen_auth("root")
        async with st.client(auth=anull) as c:
            assert (await c.get(())).value == 123

        r = await run_c("-a", "root", "auth", "user", "list")
        assert r.stdout == "*\n"

        r = await run_c("-a", "root", "auth", "user", "list", "-v")
        assert (
            r.stdout
            == """\
ident: '*'
kind: user
typ: root
"""
        )


@pytest.mark.trio
async def test_23_auth_test(autojump_clock):  # pylint: disable=unused-argument
    async with stdtest(args={"init": 123}, tocks=80) as st:
        (s,) = st.s
        h = p = None
        for h, p, *_ in s.ports:
            if h[0] != ":":
                break
        run_c = partial(run, "-D", "client", "-h", h, "-p", p)
        await run_c("data", "set", "-v", "there", "hello")

        await run_c("auth", "-m", "root", "user", "add")
        await run_c("auth", "-m", "root", "init")

        run_a = partial(run_c, "-a", "root", "auth", "-m", "_test")
        await run_a("user", "add", "name=fubar")
        res = await run_a("user", "list")
        assert res.stdout == "fubar\n"

        res = await run_a("user", "list", "-v")
        assert (
            res.stdout
            == """\
ident: fubar
kind: user
typ: _test
"""
        )
        await run_c("-a", "root", "auth", "-m", "_test", "init", "-s")

        with pytest.raises(ClientAuthMethodError):
            await run_c("-a", "root", "data", "get", "hello")
        with pytest.raises(ClientAuthRequiredError):
            await run_c("data", "get", "hello")
        with pytest.raises(jsonschema.ValidationError):
            await run_c("-a", "_test", "data", "get", "hello")

        run_t = partial(run_c, "-a", "_test name=fubar")
        res = await run_t("data", "get", "hello")
        assert res.stdout == "'there'\n"


@pytest.mark.trio
async def test_24_auth_password(autojump_clock):
    async with stdtest(args={"init": 123}, tocks=99) as st:
        (s,) = st.s
        autojump_clock.autojump_threshold = 1
        h = p = None
        for h, p, *_ in s.ports:
            if h[0] != ":":
                break
        run_c = partial(run, "-D", "client", "-h", h, "-p", p)
        await run_c("data", "set", "-v", "42", "answers.life etc:.")

        await run_c("auth", "-m", "root", "user", "add")
        await run_c("auth", "-m", "root", "init")

        run_p = partial(run_c, "-a", "root", "auth", "-m", "password")
        await run_p("user", "add", "name=joe", "password=test123")
        res = await run_p("user", "list")
        assert res.stdout == "joe\n"
        res = await run_p("user", "list", "-v")
        assert (
            res.stdout
            == """\
ident: joe
kind: user
typ: password
"""
        )
        run_u = partial(run_c, "-a", "password name=joe password=test123")
        await run_c("-a", "root", "data", "set", "-v", 42, "answers.life etc:.")
        with pytest.raises(ClientAuthMethodError):
            res = await run_u("data", "get", "answers.life etc:.")
        await run_c("-a", "root", "auth", "-m", "password", "init", "-s")
        res = await run_u("data", "get", "answers.life etc:.")
        assert res.stdout == "42\n"
        run_u = partial(run_c, "-a", "password name=joe password=test1234")
        with pytest.raises(ServerError) as se:
            res = await run_u("data", "get", "answers.life etc:.")
        assert str(se.value).startswith("AuthFailedError(")
        assert "hashes do not match" in str(se.value)



import pytest

from distkv.mock.mqtt import stdtest

from distkv.code import CodeRoot, ModuleRoot
from distkv.errors import ErrorRoot
from distkv.util import P
import logging

logger = logging.getLogger(__name__)


@pytest.mark.trio
async def test_81_basic(autojump_clock):  # pylint: disable=unused-argument
    async with stdtest(args={"init": 123}) as st:
        async with st.client() as c:
            await ErrorRoot.as_handler(c)
            cr = await CodeRoot.as_handler(c)
            await cr.add(P("forty.two"), code="return 42")
            assert cr("forty.two") == 42


@pytest.mark.trio
async def test_82_module(autojump_clock):  # pylint: disable=unused-argument
    async with stdtest(args={"init": 123}, tocks=40) as st:
        async with st.client() as c:
            await ErrorRoot.as_handler(c)
            m = await ModuleRoot.as_handler(c)
            await m.add(
                P("bar.baz"),
                code="""
def quux():
    return 42
""",
            )
            cr = await CodeRoot.as_handler(c)
            await cr.add(
                P("forty.two"),
                code="""
from bar.baz import quux
return quux()
""",
            )
            assert cr("forty.two") == 42



import pytest

from distkv.mock.mqtt import stdtest

from distkv.util import P

import logging

logger = logging.getLogger(__name__)


@pytest.mark.trio
async def test_61_basic(autojump_clock):  # pylint: disable=unused-argument
    async with stdtest(args={"init": 123}, tocks=30) as st:
        async with st.client() as c:
            async with c.mirror(P("foo"), need_wait=True) as m:
                assert (await c.get(P(":"))).value == 123

                r = await c.set(P("foo"), value="hello", nchain=3)
                r = await c.set(P("foo.bar"), value="baz", nchain=3)
                await m.wait_chain(r.chain)
                assert m.value == "hello"
                assert m["bar"].value == "baz"

                pass  # mirror end
            pass  # client end
        pass  # server end



import weakref
import jsonschema

from .model import Entry
from .util import make_proc, NotGiven, singleton, Path
from .exceptions import ClientError, ACLError


# TYPES


class MetaEntry(Entry):
    def __init__(self, *a, **k):
        super().__init__(*a, **k)
        self._metaroot = self.parent._metaroot

    @property
    def metaroot(self):
        "Stores a link to the meta root because some do need it."
        return self._metaroot()


class TypeEntry(Entry):
    """I am a type-checking node.
    """

    _code = None
    _schema = None

    def check_value(self, value, entry=None, **kv):
        self.parent.check_value(value, entry=entry, **kv)
        if self._schema is not None:
            jsonschema.validate(instance=value, schema=self._schema)
        if self._code is not None:
            self._code(value, entry=entry, data=self._data, **kv)

    async def set(self, value):
        code = None
        schema = None
        if value is not NotGiven and (
            value.get("code", None) is not None or value.get("schema", None) is not None
        ):
            schema = value.get("schema", None)
            if value.get("code", None) is None:
                code = None
            else:
                code = make_proc(value.code, ("value",), self.path)
            if len(value.good) < 2:
                raise RuntimeError("Must have check-good values")
            if not value.bad:
                raise RuntimeError("Must have check-bad values")
            for v in value["good"]:
                self.parent.check_value(v)
                try:
                    if schema is not None:
                        jsonschema.validate(instance=v, schema=schema)
                    if code is not None:
                        code(v, entry=None)
                except Exception:
                    raise ValueError(f"failed on good value {v!r}")
            for v in value["bad"]:
                self.parent.check_value(v)
                try:
                    if schema is not None:
                        jsonschema.validate(instance=v, schema=schema)
                    if code is not None:
                        code(v, entry=None)
                except Exception:
                    pass
                else:
                    raise ValueError(f"did not fail on {v!r}")

        await super().set(value)
        self._code = code
        self._schema = schema


TypeEntry.SUBTYPE = TypeEntry


class TypeRoot(Entry):
    """I am the root of DistKV's type hierarchy.
    """

    SUBTYPE = TypeEntry

    async def set(self, value):
        if value is not NotGiven:
            raise ValueError("This node can't have data.")

    def check_value(self, value, entry=None, **kv):
        pass


# MATCH (for types)


class MatchEntry(MetaEntry):
    """I represent a match from a path to a type.

    My data requires a "type" attribute that's a path in the "type"
    hierarchy that's next to my MatchRoot.
    """

    async def set(self, value):
        if value is NotGiven:
            pass
        elif isinstance(value.type, str):
            value.type = (value.type,)
        elif not isinstance(value.type, (Path, list, tuple)):
            raise ValueError("Type is not a list")
        try:
            self.metaroot["type"].follow(value.type, create=False)
        except KeyError:
            import pdb

            pdb.set_trace()
            raise ClientError("This type does not exist")
        # crashes if nonexistent
        await super().set(value)


MatchEntry.SUBTYPE = MatchEntry


class NodeFinder:
    """A generic object that can walk down a possibly-wildcard-equipped path.

    Example: given a path `one.two.three` and a root `bar` with subtree `bar.#.three`,
    `NodeFinder(bar).step(one).step(two).step(three)` will return the node
    at `bar.#.three` (assuming that nothing more specific hangs off `bar`).
    """

    def __init__(self, meta):
        if isinstance(meta, list):
            self.steps = meta
        else:
            self.steps = ((meta, False),)

    def step(self, name, new=False):
        steps = []
        for node, keep in self.steps:
            if name in node:
                steps.append((node[name], False))
            if name is None:
                continue
            if "+" in node:
                steps.append((node["+"], False))
            if "#" in node:
                steps.append((node["#"], True))
            if keep:
                steps.append((node, True))
            # Nodes found with '#' stay on the list
            # so that they can match multiple entries.
        # if not steps:
        #    raise KeyError(name)
        if new:
            return type(self)(steps, **self.copy_args)
        else:
            self.steps = steps
            return self

    @property
    def copy_args(self):
        return {}

    @property
    def result(self):
        for node, _keep in self.steps:
            if node._data is not NotGiven:
                return node
        return None


class ACLFinder(NodeFinder):
    """A NodeFinder which expects ACL strings as elements"""

    _block = ""

    @property
    def copy_args(self):
        return {"blocked": self._block}

    def __init__(self, acl, blocked=None):
        if isinstance(acl, ACLFinder):
            if blocked is None:
                blocked = acl._block
            acl = acl.steps
        super().__init__(acl)
        if blocked is not None:
            self._block = blocked

    def allows(self, x):
        if x in self._block:
            return False

        r = self.result
        if r is None:
            return True
        return x in r.data

    def block(self, c):
        if c not in self._block:
            self._block += c

    def check(self, x):
        if not self.allows(x):
            raise ACLError(self.result, x)


class ACLStepper(ACLFinder):
    """An ACLFinder which returns a copy of itself at every `step`."""

    def step(self, name, new=True):
        return super().step(name, new=new)


@singleton
class NullACL(ACLStepper):
    """This singleton represents an ACL that never checks anything."""

    result = "-"

    def __init__(self):  # pylint: disable=super-init-not-called
        pass

    def allows(self, x):
        return x != "a"

    def check(self, x):
        return

    def block(self, c):
        pass

    def step(self, name, new=None):
        return self


class MetaPathEntry(MetaEntry):
    def _find_node(self, entry):
        """Search for the most-specific match.

        Match entries whose values are missing are not considered.
        """
        f = NodeFinder(self)
        for n in entry.path:
            f.step(n)
        return f.result


class MatchRoot(MetaPathEntry):
    """I am the root of DistKV's type hierarchy.
    """

    SUBTYPE = MatchEntry

    async def set(self, value):
        if value is not None:
            raise ValueError("This node can't have data.")

    def check_value(self, value, entry, **kv):
        """Check this value for this entry against my match hierarchy"""
        match = self._find_node(entry)
        if match is None:
            return
        typ = self.parent["type"].follow(match._data["type"])
        return typ.check_value(value, entry=entry, match=match, **kv)


class CodecEntry(Entry):
    """I am a codec.
    """

    _enc = None
    _dec = None

    def enc_value(self, value, entry=None, **kv):
        if self._enc is not None:
            try:
                value = self._enc(value, entry=entry, data=self._data, **kv)
            except TypeError:
                if value is not None:
                    raise
        return value

    def dec_value(self, value, entry=None, **kv):
        if self._dec is not None:
            try:
                value = self._dec(value, entry=entry, data=self._data, **kv)
            except TypeError:
                if value is not None:
                    raise
        return value

    async def set(self, value):
        enc = None
        dec = None
        if value is not None and value.decode is not None:
            if not value["in"]:
                raise RuntimeError("Must have tests for decoding")
            dec = make_proc(value.decode, ("value",), self.path)
            for v, w in value["in"]:
                try:
                    r = dec(v)
                except Exception as exc:
                    raise ValueError(
                        f"failed decoder at {self.path} on {v!r} with {exc!r}"
                    ) from exc
                else:
                    if r != w:
                        raise ValueError(f"Decoding at {self.path}: {v!r} got {r!r}, not {w!r}")

        if value is not None and value.encode is not None:
            if not value["out"]:
                raise RuntimeError("Must have tests for encoding")
            enc = make_proc(value.encode, ("value",), self.path)
            for v, w in value["out"]:
                try:
                    r = enc(v)
                except Exception as exc:
                    raise ValueError(
                        f"failed encoder at {self.path} on {v!r} with {exc!r}"
                    ) from exc
                else:
                    if r != w:
                        raise ValueError(f"Encoding at {self.path}: {v!r} got {r!r}, not {w!r}")

        await super().set(value)
        self._enc = enc
        self._dec = dec


CodecEntry.SUBTYPE = CodecEntry


class CodecRoot(Entry):
    """I am the root of DistKV's codec hierarchy.
    """

    SUBTYPE = CodecEntry

    async def set(self, value):
        if value is not None:
            raise ValueError("This node can't have data.")

    def check_value(self, value, entry=None, **kv):
        pass


# CONV (for codecs)


class ConvEntry(MetaEntry):
    """I represent a converter from a path to a type.

    My data requires a "codec" attribute that's a path in the "codecs"
    hierarchy that's next to my Root.
    """

    async def set(self, value):
        if value is not NotGiven:
            # can't run a nonexistent value through a codec

            if not hasattr(value, "codec"):
                raise ValueError("Duh? " + repr(value))
            if isinstance(value.codec, str):
                value.codec = (value.codec,)
            elif not isinstance(value.codec, (list, tuple)):
                raise ValueError("Codec is not a string or list")
            try:
                self.metaroot["codec"].follow(value.codec, create=False)
            except KeyError:
                raise ClientError("This codec does not exist")
        await super().set(value)


ConvEntry.SUBTYPE = ConvEntry


class ConvNull:
    """I am a dummy translator"""

    @staticmethod
    def enc_value(value, **_kw):
        return value

    @staticmethod
    def dec_value(value, **_kw):
        return value


ConvNull = ConvNull()


class ConvName(MetaPathEntry):
    """I am a named tree for conversion entries.
    """

    SUBTYPE = ConvEntry

    async def set(self, value):
        if value is not None:
            raise ValueError("This node can't have data.")

    def enc_value(self, value, entry, **kv):
        """Check this value for this entry against my converter hierarchy"""
        conv = self._find_node(entry)
        if conv is None:
            return value
        codec = self.metaroot["codec"].follow(conv._data["codec"])
        return codec.enc_value(value, entry=entry, conv=conv, **kv)

    def dec_value(self, value, entry, **kv):
        """Check this value for this entry against my converter hierarchy"""
        conv = self._find_node(entry)
        if conv is None:
            return value
        codec = self.metaroot["codec"].follow(conv._data["codec"])
        return codec.dec_value(value, entry=entry, conv=conv, **kv)


class ConvRoot(MetaEntry):
    SUBTYPE = ConvName

    async def set(self, value):
        if value is not None:
            raise ValueError("This node can't have data.")


class AclEntry(MetaPathEntry):
    """I represent an ACL.

    My data must be a string of ACLish characters.
    """

    async def set(self, value):
        if value is NotGiven:
            pass
        elif not isinstance(value, str):
            raise ValueError("ACL is not a string")
        await super().set(value)


AclEntry.SUBTYPE = AclEntry


class AclName(AclEntry):
    """I am a named tree for ACL entries.
    """

    SUBTYPE = AclEntry

    async def check(self, entry, typ):
        acl = self._find_node(entry)
        if acl is None:
            return None
        return typ in acl.value


class AclRoot(MetaEntry):
    SUBTYPE = AclName

    async def set(self, value):
        if value is not None:
            raise ValueError("This node can't have data.")


class DelRoot(Entry):
    SUBTYPE = None

    async def set(self, value):
        await super().set(value)
        await self.root.server.del_check(value)


class ActorRoot(Entry):
    SUBTYPE = None
    SUBTYPES = {"del": DelRoot}


# ROOT


class MetaRootEntry(Entry):  # not MetaEntry
    """I am the special node off the DistKV root that's named ``None``."""

    SUBTYPES = {
        "type": TypeRoot,
        "match": MatchRoot,
        "acl": AclRoot,
        "codec": CodecRoot,
        "conv": ConvRoot,
        "actor": ActorRoot,
    }

    def __init__(self, *a, **k):
        super().__init__(*a, **k)
        self._metaroot = weakref.ref(self)


Entry.SUBTYPES[None] = MetaRootEntry


class RootEntry(Entry):
    """I am the root of the DistKV data tree."""

    def __init__(self, server, *a, **k):
        super().__init__("ROOT", None, *a, **k)
        self._server = weakref.ref(server)

    @property
    def server(self):
        return self._server()

    SUBTYPES = {None: MetaRootEntry}



"""
Client code.

Main entry point: :func:`open_client`.
"""

import anyio
import outcome
import socket
import os
from typing import Tuple
from asyncscope import scope, Scope, main_scope

rs = os.environ.get("PYTHONHASHSEED", None)
if rs is None:
    import random
else:  # pragma: no cover
    import trio._core._run as tcr

    random = tcr._r

try:
    from contextlib import asynccontextmanager, AsyncExitStack
except ImportError:  # pragma: no cover
    from async_generator import asynccontextmanager
    from async_exit_stack import AsyncExitStack

from .util import (
    attrdict,
    gen_ssl,
    num2byte,
    byte2num,
    PathLongener,
    NotGiven,
    combine_dict,
    ValueEvent,
    Path,
)
from .default import CFG
from .exceptions import (
    ClientAuthMethodError,
    ClientAuthRequiredError,
    ServerClosedError,
    ServerConnectionError,
    ServerError,
    error_types,
    CancelledError,
)
from .codec import packer, stream_unpacker

import logging

logger = logging.getLogger(__name__)

__all__ = ["NoData", "ManyData", "open_client", "client_scope", "StreamedRequest"]


class NoData(ValueError):
    """No reply arrived"""


class ManyData(ValueError):
    """More than one reply arrived"""


@asynccontextmanager
async def open_client(_main_name="_distkv_client", **cfg):
    """
    This async context manager returns an opened client connection.

    There is no attempt to reconnect if the connection should fail.

    The client connection is run within a separate `asyncscope.ScopeSet`.
    If you're already using asyncscope in your code, you should use
    `client_scope` instead.
    """
    async with main_scope(name=_main_name):
        client = await client_scope(**cfg)
        yield client
        pass  # end _connected
    pass  # end client


async def client_scope(**cfg):
    """
    Return the opened client connection, by way of an asyncscope service.

    The configuration's 'connect' dict may include a name to disambiguate
    multiple connections. This name must not be equal to your main code's.

    There is no attempt to reconnect if the connection should fail.
    """

    async def _mgr(cfg):
        client = Client(cfg)
        async with client._connected() as client:
            await scope.register(client)
            await scope.no_more_dependents()
            pass  # end _connected

    name = cfg["connect"].get("name", None)
    if name is None:
        name = "_distkv_conn"  # MUST NOT be the same as in open_client
    res = await scope.service(name, _mgr, cfg)
    return res


class StreamedRequest:
    """
    This class represents a bidirectional multi-message request.

    stream: True if you need to send a multi-message request.
            Set to None if you already sent a single-message request.
    report_start: True if the initial state=start message of a multi-reply
                  should be included in the iterator.
                  If False, the message is available as ``.start_msg``.
    TODO: add rate limit.

    Call ``.send(**params)`` to send something; call ``.recv()``
    or async-iterate for receiving.
    """

    start_msg = None
    end_msg = None

    def __init__(self, client, seq, stream: bool = False, report_start: bool = False):
        self._stream = stream
        self._client = client
        self.seq = seq
        self._stream = stream
        self.q = anyio.create_queue(100)
        self._client._handlers[seq] = self
        self._reply_stream = None
        self.n_msg = 0
        self._report_start = report_start
        self._started = anyio.create_event()
        self._path_long = lambda x: x
        # None: no message yet; True: begin seen; False: end or single message seen

    async def set(self, msg):
        """Called by the read loop to process a command's result"""
        self.n_msg += 1
        if "error" in msg:
            logger.info("ErrorMsg: %s", msg)
            if self.q is not None:
                try:
                    cls = error_types[msg["etype"]]
                except KeyError:
                    cls = ServerError
                await self.q.put(outcome.Error(cls(msg.error)))
            return
        state = msg.get("state", "")

        if state == "start":
            if self._reply_stream is not None:  # pragma: no cover
                raise RuntimeError("Recv state 2", self._reply_stream, msg)
            self._reply_stream = True
            self.start_msg = msg
            await self._started.set()
            if self._report_start:
                await self.q.put(outcome.Value(msg))

        elif state == "end":
            if self._reply_stream is not True:  # pragma: no cover
                raise RuntimeError("Recv state 3", self._reply_stream, msg)
            self._reply_stream = None
            self.end_msg = msg
            if self.q is not None:
                await self.q.put(None)
            return False

        else:
            if state not in ("", "uptodate"):  # pragma: no cover
                logger.warning("Unknown state: %s", msg)

            if self._reply_stream is False:  # pragma: no cover
                raise RuntimeError("Recv state 1", self._reply_stream, msg)
            elif self._reply_stream is None:
                self._reply_stream = False
            if self.q is not None:
                await self.q.put(outcome.Value(msg))
                if self._reply_stream is False:
                    await self.q.put(None)

    async def get(self):
        """Receive a single reply"""
        pass  # receive reply
        if self._reply_stream:
            raise RuntimeError("Unexpected multi stream msg")
        msg = await self.recv()
        if self._reply_stream or self.n_msg != 1:
            raise RuntimeError("Unexpected multi stream msg")
        return msg

    def __iter__(self):
        raise RuntimeError("You need to use 'async for …'")

    __next__ = __iter__

    def __aiter__(self):
        return self

    async def __anext__(self):
        res = await self.q.get()
        if res is None:
            self.q = None  # prevent deadlock if called again
            raise StopAsyncIteration
        try:
            res = res.unwrap()
        except CancelledError:
            raise StopAsyncIteration  # just terminate
        self._path_long(res)
        logger.debug("OneResult: %s", res)
        return res

    async def send(self, **params):
        # logger.debug("Send %s", params)
        if not self._stream:
            if self._stream is None:
                raise RuntimeError("You can't send more than one request")
            self._stream = None
        elif self._stream is True:
            self._stream = 2
            params["state"] = "start"
        elif self._stream == 2 and params.get("state", "") == "end":
            self._stream = None
        logger.debug("Send %s", {"seq": self.seq, **params})
        await self._client._send(seq=self.seq, **params)

    async def recv(self):
        return await self.__anext__()

    async def cancel(self):
        if self.q is not None:
            await self.q.put(outcome.Error(CancelledError()))
        await self.aclose()

    async def wait_started(self):
        await self._started.wait()

    async def aclose(self, timeout=0.2):
        if self.q is not None:
            await self.q.put(None)
        if self._stream == 2:
            await self._client._send(seq=self.seq, state="end")
            if timeout is not None:
                async with anyio.move_on_after(timeout):
                    try:
                        await self.recv()
                    except StopAsyncIteration:
                        return
            req = await self._client._request(action="stop", task=self.seq, _async=True)
            return await req.get()


class _SingleReply:
    """
    This class represents a single-message reply.
    It will delegate itself to a StreamedRequest if a multi message reply
    arrives.
    """

    def __init__(self, conn, seq, params):
        self._conn = conn
        self.seq = seq
        self.q = ValueEvent()
        self._params = params

    async def set(self, msg):
        """Called by the read loop to process a command's result"""
        if msg.get("state") == "start":
            res = StreamedRequest(self._conn, self.seq, stream=None)
            await res.set(msg)
            await self.q.set(res)
            return res
        elif "error" in msg:
            msg["request_params"] = self._params
            logger.info("ErrorMsg: %s", msg)
            await self.q.set_error(ServerError(msg.error))
        else:
            await self.q.set(msg)
        return False

    async def get(self):
        """Wait for and return the result.

        This is a coroutine.
        """
        return await self.q.get()

    async def cancel(self):
        pass


class ClientConfig:
    """Accessor for configuration, possibly stored in DistKV.
    """

    _changed = None  # pylint

    def __init__(self, client, *a, **k):  # pylint: disable=unused-argument
        self._init(client)

    def _init(self, client):
        self._client = client
        self._current = {}
        self._changed = anyio.create_event()

    def __getattr__(self, k):
        if k.startswith("_"):
            return object.__getattribute__(self, k)
        v = self._current.get(k, NotGiven)
        if v is NotGiven:
            try:
                v = self._client._cfg[k]
            except KeyError:
                raise AttributeError(k) from None
        return v

    def __contains__(self, k):
        return k in self._current or k in self._client._cfg

    async def _update(self, k, v):
        """
        Update this config entry. The new data is combined with the static
        configuration; the old data is discarded.
        """
        self._current[k] = combine_dict(v, self._client._cfg.get(k, {}))
        c, self._changed = self._changed, anyio.create_event()
        await c.set()

    async def _watch(self):
        class CfgWatcher:
            def __ainit__(slf):  # pylint: disable=no-self-argument
                return slf

            async def __anext__(slf):  # pylint: disable=no-self-argument
                await self._changed.wait()

        return CfgWatcher()


class Client:
    """
    The client side of a DistKV connection.

    Use `open_client` or `client_scope` to use this class.
    """

    _server_init = None  # Server greeting
    _dh_key = None
    _config = None
    _socket = None
    tg: anyio.abc.TaskGroup = None
    scope: Scope = None
    _n = None
    exit_stack = None

    server_name = None
    client_name = None

    def __init__(self, cfg: dict):
        self._cfg = combine_dict(cfg, CFG, cls=attrdict)
        self.config = ClientConfig(self)

        self._seq = 0
        self._handlers = {}
        self._send_lock = anyio.create_lock()
        self._helpers = {}
        self._name = "".join(random.choices("abcdefghjkmnopqrstuvwxyz23456789", k=9))

    @property
    def name(self):
        return self._name

    @property
    def node(self):
        return self._server_init["node"]

    async def get_tock(self):
        """Fetch the next tock value from the server."""
        m = await self._request("get_tock")
        return m.tock

    async def unique_helper(self, path, factory):
        """
        Run a (single) async context manager on that path.
        """
        p = str(Path.build(path))

        async def with_factory(f):
            async with f() as r:
                await scope.register(r)
                await scope.no_more_dependents()

        return await scope.service(p, with_factory, factory)

    async def _handle_msg(self, msg):
        try:
            seq = msg.seq
        except AttributeError:
            if "error" in msg:
                raise RuntimeError("Server error", msg.error)
            raise RuntimeError("Reader got out of sync: " + str(msg))
        try:
            hdl = self._handlers[seq]
        except KeyError:
            logger.warning("Spurious message %s: %s", seq, msg)
            return

        res = await hdl.set(msg)
        if res is False:
            del self._handlers[seq]
        elif res:
            self._handlers[seq] = res

    async def dh_secret(self, length=1024):
        """Exchange a diffie-hellman secret with the server"""
        if self._dh_key is None:
            from diffiehellman.diffiehellman import DiffieHellman

            def gen_key():
                k = DiffieHellman(key_length=length, group=(5 if length < 32 else 14))
                k.generate_public_key()
                return k

            k = await anyio.run_in_thread(gen_key)
            res = await self._request(
                "diffie_hellman", pubkey=num2byte(k.public_key), length=length
            )  # length=k.key_length
            await anyio.run_in_thread(k.generate_shared_secret, byte2num(res.pubkey))
            self._dh_key = num2byte(k.shared_secret)[0:32]
        return self._dh_key

    async def _send(self, **params):
        async with self._send_lock:
            sock = self._socket
            if sock is None:
                raise ServerClosedError("Disconnected")

            try:
                p = packer(params)
            except TypeError as e:
                raise ValueError(f"Unable to pack: {params!r}") from e
            await sock.send_all(p)

    async def _reader(self, *, evt=None):
        """Main loop for reading
        """
        unpacker = stream_unpacker()

        async with anyio.open_cancel_scope():
            # XXX store the scope so that the redaer may get cancelled?
            if evt is not None:
                await evt.set()
            try:
                while True:
                    for msg in unpacker:
                        # logger.debug("Recv %s", msg)
                        try:
                            await self._handle_msg(msg)
                        except anyio.exceptions.ClosedResourceError:
                            raise RuntimeError(msg)

                    if self._socket is None:
                        break
                    try:
                        buf = await self._socket.receive_some(4096)
                    except anyio.exceptions.ClosedResourceError:
                        return  # closed by us
                    if len(buf) == 0:  # Connection was closed.
                        raise ServerClosedError("Connection closed by peer")
                    unpacker.feed(buf)

            finally:
                async with anyio.fail_after(2, shield=True):
                    hdl, self._handlers = self._handlers, None
                    for m in hdl.values():
                        try:
                            await m.cancel()
                        except anyio.exceptions.ClosedResourceError:
                            pass

    async def _request(
        self, action, iter=None, seq=None, _async=False, **params
    ):  # pylint: disable=redefined-builtin  # iter
        """Send a request. Wait for a reply.

        Args:
          action (str): what to do. If ``seq`` is set, this is the stream's
            state, which should be ``None`` or ``'end'``.
          seq: Sequence number to use. Only when terminating a
            multi-message request.
          _async: don't wait for a reply (internal!)
          params: whatever other data the action needs
          iter: A flag how to treat multi-line replies.
            ``True``: always return an iterator
            ``False``: Never return an iterator, raise an error
                       if no or more than on reply arrives
            Default: ``None``: return a StreamedRequest if multi-line
                                otherwise return directly

        Any other keywords are forwarded to the server.
        """
        if self._handlers is None:
            raise anyio.exceptions.ClosedResourceError()
        if seq is None:
            act = "action"
            self._seq += 1
            seq = self._seq
        else:
            act = "state"

        if action is not None:
            params[act] = action
        params["seq"] = seq
        res = _SingleReply(self, seq, params)
        self._handlers[seq] = res

        logger.debug("Send %s", params)
        await self._send(**params)
        if _async:
            return res

        res = await res.get()
        logger.debug("Result %s", res)

        if iter is True and not isinstance(res, StreamedRequest):

            async def send_one(res):
                yield res

            res = send_one(res)

        elif iter is False and isinstance(res, StreamedRequest):
            rr = None
            async for r in res:
                if rr is not None:
                    raise ManyData(action)
                rr = r
            if rr is None:
                raise NoData(action)
            res = rr
        return res

    @asynccontextmanager
    async def _stream(self, action, stream=False, **params):
        """Send and receive a multi-message request.

        Args:
          ``action``: what to do
          ``params``: whatever other data the action needs
          ``stream``: whether to enable multi-line requests
                      via ``await stream.send(**params)``

        This is a context manager. Use it like this::

            async with client._stream("update", path=P("private.storage"),
                    stream=True) as req:
                with MsgReader("/tmp/msgs.pack") as f:
                    for msg in f:
                        await req.send(msg)
            # … or …
            async with client._stream("get_tree", path=P("private.storage)) as req:
                for msg in req:
                    await process_entry(msg)
            # … or maybe … (auth does this)
            async with client._stream("interactive_thing", path=P(':n.foo)) as req:
                msg = await req.recv()
                while msg.get(s,"")=="more":
                    await foo.send(s="more",value="some data")
                    msg = await req.recv()
                await foo.send(s="that's all then")

        Any server-side exception will be raised on recv.

        The server-side command will be killed if you leave the loop
        without having read a "state=end" message.
        """
        self._seq += 1
        seq = self._seq

        # logger.debug("Send %s", params)
        if self._handlers is None:
            raise anyio.exceptions.ClosedResourceError("Closed already")
        res = StreamedRequest(self, seq, stream=stream)
        if "path" in params and params.get("long_path", False):
            res._path_long = PathLongener(params["path"])
        await res.send(action=action, **params)
        await res.wait_started()
        try:
            yield res
        except BaseException as exc:
            if stream:
                await res.send(error=repr(exc))
            raise
        finally:
            async with anyio.fail_after(2, shield=True):
                await res.aclose()

    async def _run_auth(self, auth=None):
        hello = self._server_init
        sa = hello.get("auth", ())
        if not sa or not sa[0]:
            # no auth required
            if auth:
                logger.info("Tried to use auth=%s, but not required.", auth._auth_method)
            return
        if not auth:
            raise ClientAuthRequiredError("You need to log in using:", sa[0])
        if auth._auth_method != sa[0]:
            raise ClientAuthMethodError(f"You cannot use {auth._auth_method!r} auth", sa)
        if getattr(auth, "_DEBUG", False):
            auth._length = 16
        await auth.auth(self)

    @asynccontextmanager
    async def _connected(self):
        """
        This async context manager handles the actual TCP connection to
        the DistKV server.
        """
        hello = ValueEvent()
        self._handlers[0] = hello

        cfg = self._cfg["connect"]
        host = cfg["host"]
        port = cfg["port"]
        auth = cfg["auth"]
        if auth is not None:
            from .auth import gen_auth

            auth = gen_auth(auth)
        init_timeout = cfg["init_timeout"]
        ssl = gen_ssl(cfg["ssl"], server=False)

        # logger.debug("Conn %s %s",self.host,self.port)
        try:
            ctx = await anyio.connect_tcp(host, port, ssl_context=ssl, autostart_tls=False)
        except socket.gaierror:
            raise ServerConnectionError(host, port)

        else:
            async with ctx as stream, AsyncExitStack() as ex:
                self.scope = scope.get()
                # self.tg = tg  # TODO might not be necessary
                self.exit_stack = ex

                if ssl:
                    await stream.start_tls()
                try:
                    self._socket = stream
                    await self.scope.spawn(self._reader)
                    async with anyio.fail_after(init_timeout):
                        self._server_init = await hello.get()
                        logger.debug("Hello %s", self._server_init)
                        self.server_name = self._server_init.node
                        self.client_name = cfg["name"] or self.server_name
                        await self._run_auth(auth)

                    from .config import ConfigRoot

                    self._config = await ConfigRoot.as_handler(self, require_client=False)

                except TimeoutError:
                    raise
                except socket.error as e:
                    raise ServerConnectionError(host, port) from e
                else:
                    yield self
                finally:
                    async with anyio.fail_after(2, shield=True):
                        # Clean up our hacked config
                        try:
                            del self._config
                        except AttributeError:
                            pass
                        self.config = ClientConfig(self)

                    # await self.tg.cancel_scope.cancel()
        finally:
            self._socket = None
            self.tg = None

    # externally visible interface ##########################

    def get(self, path, *, nchain=0):
        """
        Retrieve the data at a particular subtree position.

        Usage::
            res = await client.get("foo","bar")

        If you want to update this value, you should retrieve its change chain entry
        so that a competing update can be detected::

            res = await client.get("foo","bar", nchain=-1)
            res = await client.set("foo","bar", value=res.value+1, chain=res.chain)

        For lower overhead and set-directly-after-get change, nchain may be 1 or 2.
        """
        if isinstance(path, str):
            raise RuntimeError("You need a path, not a string")
        return self._request(action="get_value", path=path, iter=False, nchain=nchain)

    def set(self, path, value=NotGiven, *, chain=NotGiven, prev=NotGiven, nchain=0, idem=None):
        """
        Set or update a value.

        Usage::
            await client.set("foo","bar", value="baz", chain=None)

        Arguments:
            value: the value to set. Duh. ;-)
            chain: the previous value's change chain. Use ``None`` for new values.
            prev: the previous value. Discouraged; use ``chain`` instead.
            nchain: set to retrieve the node's chain tag, for further updates.
            idem: if True, no-op if the value doesn't change
        """
        if isinstance(path, str):
            raise RuntimeError("You need a path, not a string")
        if value is NotGiven:
            raise RuntimeError("You need to supply a value, or call 'delete'")

        kw = {}
        if prev is not NotGiven:
            kw["prev"] = prev
        if chain is not NotGiven:
            kw["chain"] = chain
        if idem is not None:
            kw["idem"] = idem

        return self._request(
            action="set_value", path=path, value=value, iter=False, nchain=nchain, **kw
        )

    def delete(self, path, *, chain=NotGiven, prev=NotGiven, nchain=0):
        """
        Delete a node.

        Usage::
            await client.delete("foo","bar")

        Arguments:
            chain: the previous value's change chain.
            prev: the previous value. Discouraged; use ``chain`` instead.
            nchain: set to retrieve the node's chain, for setting a new value.
        """
        if isinstance(path, str):
            raise RuntimeError("You need a path, not a string")
        kw = {}
        if prev is not NotGiven:
            kw["prev"] = prev
        if chain is not NotGiven:
            kw["chain"] = chain

        return self._request(action="delete_value", path=path, iter=False, nchain=nchain, **kw)

    async def list(self, path, *, with_data=False, empty=None, **kw):
        """
        Retrieve the next data level.

        Args:
          with_data (bool): Return the data along with the keys. Default False.
          empty (bool): Return [names of] empty nodes. Default True if
            with_data is not set.
        """
        if isinstance(path, str):
            raise RuntimeError("You need a path, not a string")
        if empty is None:
            empty = not with_data
        res = await self._request(action="enum", path=path, with_data=with_data, empty=empty, **kw)
        return res.result

    async def get_tree(self, path, *, long_path=True, **kw):
        """
        Retrieve a complete DistKV subtree.

        This call results in a stream of tree nodes. Storage of these nodes,
        if required, is up to the caller. Also, the server does not
        take a snapshot for you, thus the data may be inconsistent.

        Use :meth:`mirror` if you want this tree to be kept up-to-date or
        if you need a consistent snapshot.

        Args:
          nchain (int): Length of change chain to add to the results, for updating.
          min_depth (int): min level of nodes to retrieve.
          max_depth (int): max level of nodes to retrieve.
          long_path (bool): if set (the default), pass the result through PathLongener

        """
        if isinstance(path, str):
            raise RuntimeError("You need a path, not a string")
        if long_path:
            lp = PathLongener()
        async for r in await self._request(
            action="get_tree", path=path, iter=True, long_path=True, **kw
        ):
            if long_path:
                lp(r)
            yield r

    def delete_tree(self, path, *, nchain=0):
        """
        Delete a whole subtree.

        If you set ``nchain``, this call will return an async iterator over
        the deleted nodes; if not, the single return value only contains the
        number of deleted nodes.
        """
        if isinstance(path, str):
            raise RuntimeError("You need a path, not a string")
        return self._request(action="delete_tree", path=path, nchain=nchain)

    def stop(self, seq: int):
        """End this stream or request.

        Args:
            seq: the sequence number of the request in question.

        TODO: DistKV doesn't do per-command flow control yet, so you should
        call this method from a different task if you don't want to risk a
        deadlock.
        """
        return self._request(action="stop", task=seq)

    def watch(self, path, *, long_path=True, **kw):
        """
        Return an async iterator of changes to a subtree.

        Args:
          fetch (bool): if ``True``, also send the currect state. Be aware
            that this may overlap with processing changes: you may get
            updates before the current state is completely transmitted.
          nchain: add the nodes' change chains.
          min_depth (int): min level of nodes to retrieve.
          max_depth (int): max level of nodes to retrieve.

        The result should be passed through a :class:`distkv.util.PathLongener`.

        If ``fetch`` is set, a ``state="uptodate"`` message will be sent
        as soon as sending the current state is completed.

        DistKV will not send stale data, so you may always replace a path's
        old cached state with the newly-arrived data.
        """
        if isinstance(path, str):
            raise RuntimeError("You need a path, not a string")
        return self._stream(action="watch", path=path, iter=True, long_path=long_path, **kw)

    def mirror(self, path, *, root_type=None, **kw):
        """An async context manager that affords an update-able mirror
        of part of a DistKV store.

        Arguments:
          root_type (type): The class to use for the root. Must be
            :class:`ClientRoot` or a subclass.

        Returns: the root of this tree.

        Usage::
            async with distkv.open_client() as c:
                async with c.mirror("foo", "bar", need_wait=True) as foobar:
                    r = await c.set_value("foo", "bar", "baz", value="test")
                    await foobar.wait_chain(r.chain)
                    assert foobar['baz'].value == "test"
                pass
                # At this point you can still access the tree's data
                # via ``foobar``, but they will no longer be kept up-to-date.

        """
        if isinstance(path, str):
            raise RuntimeError("You need a path, not a string")
        if root_type is None:
            from .obj import ClientRoot

            root_type = ClientRoot
        root = root_type(self, path, **kw)
        return root.run()

    def msg_monitor(self, topic: Tuple[str], raw: bool = False):
        """
        Return an async iterator of tunneled Serf messages. This receives
        all messages sent using :meth:`msg_send` with the same tag.

        Args:
            tag: the topic / "user:" tag to monitor.
                 The first character may not be '+'.
                 Do not include Serf's "user:" prefix.
            raw: If ``True``, will not try to msgpack-decode incoming
                 messages.

        Returns: a dict.
            data: decoded data. Not present when ``raw`` is set or the
                  decoder raised an exception.
            raw: un-decoded data. Not present when '`raw`` is not set and
                 decoding succeeded.
            error: Error message. Not present when ``raw`` is set or
                   ``data`` is present.
        usage::
            async with client.msg_monitor(["test"]) as cl:
                async for msg in cl:
                    if 'error' in msg:
                        raise RuntimeError(msg.error)
                    await process_test(msg.data)
        """
        return self._stream(action="msg_monitor", topic=topic, raw=raw)

    def msg_send(self, topic: Tuple[str], data=None, raw: bytes = None):
        """
        Tunnel a user-tagged message through Serf. This sends the message
        to all active callers of :meth:`msg_monitor` which use the same tag.

        Args:
            tag: the "user:" tag to send to.
                 The first character may not be '+'.
                 Do not include Serf's "user:" prefix.
            data: to-be-encoded data (anything ``msgpack`` can process).
            raw: raw binary data to send, mutually exclusive with ``data``.
        """
        if raw is None:
            return self._request(action="msg_send", topic=topic, data=data)
        else:
            return self._request(action="msg_send", topic=topic, raw=raw)



import pytest
import trio
import anyio
import time

from distkv.mock.mqtt import stdtest

from distkv.code import CodeRoot
from distkv.runner import AllRunnerRoot
from distkv.errors import ErrorRoot
from distkv.util import P
import logging

logger = logging.getLogger(__name__)


@pytest.mark.trio
async def test_83_run(autojump_clock):  # pylint: disable=unused-argument
    async with stdtest(args={"init": 123}, tocks=50) as st:
        async with st.client() as c:
            await ErrorRoot.as_handler(c)
            cr = await CodeRoot.as_handler(c)
            r = await AllRunnerRoot.as_handler(c, subpath=(), code=cr)
            c._test_evt = anyio.create_event()
            await cr.add(
                P("forty.two"),
                code="""\
                import trio
                c=_client
                await c._test_evt.set()
                await trio.sleep(10)
                return 42
                """,
                is_async=True,
            )
            ru = r.follow(P("foo.test"), create=True)
            ru.code = P("forty.two")
            await ru.run_at(time.time() + 1)
            logger.info("Start sleep")
            with trio.fail_after(60):
                await c._test_evt.wait()
            await st.run("data get -rd_ :", do_stdout=False)
            await trio.sleep(11)

            logger.info("End sleep")

            r = await st.run("data get :")
            assert r.stdout == "123\n"

            rs = ru.state
            assert rs.started > 0
            assert rs.stopped > 0
            assert rs.backoff == 0
            assert rs.result == 42



from setuptools import setup, find_packages

LONG_DESC = open("README.rst").read()

setup(
    name="distkv",
    use_scm_version={"version_scheme": "guess-next-dev", "local_scheme": "dirty-tag"},
    description="A distributed no-master key-value store",
    url="https://github.com/smurfix/distkv",
    long_description=LONG_DESC,
    author="Matthias Urlichs",
    author_email="matthias@urlichs.de",
    license="MIT -or- Apache License 2.0",
    packages=find_packages() + ["distkv_ext.dummy"],
    # namespace_packages=["distkv_ext.dummy"],
    setup_requires=["setuptools_scm", "pytest-runner", "trustme >= 0.5"],
    install_requires=[
        "asyncclick",
        "trio >= 0.15",
        "anyio",
        "range_set >= 0.2",
        "attrs >= 19",
        "asyncserf >= 0.16",
        "asyncactor >= 0.20.5",
        "asyncscope >= 0.4.0",
        "jsonschema >= 2.5",
        "ruamel.yaml >= 0.16",
        # "argon2 >= 18.3",
        "PyNaCl >= 1.3",
        "diffiehellman",
        "psutil",
        "systemd-python",  # OWCH NO
        "simpleeval >= 0.9.10",
    ],
    tests_require=["trustme >= 0.5", "pytest", "flake8 >= 3.7", "distmqtt >= 0.30"],
    keywords=["async", "key-values", "distributed"],
    python_requires=">=3.7",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: MIT License",
        "License :: OSI Approved :: Apache Software License",
        "Framework :: AsyncIO",
        "Framework :: Trio",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Database",
        "Topic :: Home Automation",
        "Topic :: System :: Distributed Computing",
    ],
    entry_points="""
    [console_scripts]
    distkv = distkv.command:cmd
    """,
    zip_safe=True,
)



"""
Code to load external modules
"""

import os
import asyncclick as click


def load_one(name, path, endpoint=None, **ns):
    fn = os.path.join(path, name) + ".py"
    if not os.path.isfile(fn):
        fn = os.path.join(path, name, "__init__.py")
    with open(fn) as f:
        ns["__file__"] = fn

        code = compile(f.read(), fn, "exec")
        eval(code, ns, ns)  # pylint: disable=eval-used
    if endpoint is not None:
        ns = ns[endpoint]
    return ns


def _namespaces():
    import pkgutil
    import distkv_ext as ext

    return pkgutil.iter_modules(ext.__path__, ext.__name__ + ".")


_ext_cache = {}


def _cache_ext():
    """List external modules

    Yields (name,path) tuples.

    TODO: This is not zip safe.
    """
    for finder, name, ispkg in _namespaces():
        if not ispkg:
            continue
        x = name.rsplit(".", 1)[-1]
        f = os.path.join(finder.path, x)
        _ext_cache[x] = f


def list_ext(func=None):
    """List external modules

    Yields (name,path) tuples.

    TODO: This is not zip safe.
    """
    if not _ext_cache:
        _cache_ext()
    if func is None:
        yield from iter(_ext_cache.items())
        return
    for x, f in _ext_cache.items():
        fn = os.path.join(f, func) + ".py"
        if not os.path.exists(fn):
            fn = os.path.join(f, func, "__init__.py")
            if not os.path.exists(fn):
                continue
        yield (x, f)


def load_ext(name, func, endpoint=None, **kw):
    """
    Load an external module.

    Example: ``load_ext("owfs","model")`` loads …/distkw_ext/owfs/model.py
    and returns its global dict. When "ep" is given it returns the entry
    point.

    Any additional keywords are added to the module dictionary.

    TODO: This is not zip safe. It also doesn't return a proper module.
    Don't use this with modules that are also loaded the regular way.
    """

    if not _ext_cache:
        _cache_ext()

    try:
        return load_one(func, _ext_cache[name], endpoint, **kw)
    except KeyError:
        raise click.UsageError(f"I do not know {name !r}")



# command line interface

import os
import sys
import asyncclick as click
from functools import partial

from distkv.util import (
    attrdict,
    combine_dict,
    NotGiven,
    res_get,
    res_update,
    res_delete,
    yload,
    yprint,
)
from distkv.default import CFG
from distkv.ext import load_one, list_ext, load_ext
from distkv.exceptions import ClientError, ServerError

import logging
from logging.config import dictConfig

logger = logging.getLogger(__name__)


class Loader(click.Group):
    """
    A Group that can load additional commands from a subfolder.

    Caller:

        from distkv.command import Loader
        from functools import partial

        @click.command(cls=partial(Loader,__file__,'command'))
        async def cmd()
            print("I am the main program")

    Sub-Command Usage (``main`` is defined for you), e.g. in ``command/subcmd.py``::

        from distkv.command import Loader
        from functools import partial

        @main.command / group()
        async def cmd(self):
            print("I am", self.name)  # prints "subcmd"
    """

    def __init__(self, current_file, plugin, *a, **kw):
        self.__plugin = plugin
        self.__plugin_folder = os.path.dirname(current_file)
        super().__init__(*a, **kw)

    def list_commands(self, ctx):
        rv = super().list_commands(ctx)

        for filename in os.listdir(self.__plugin_folder):
            if filename[0] in "._":
                continue
            if filename.endswith(".py"):
                rv.append(filename[:-3])
            elif os.path.isfile(os.path.join(self.__plugin_folder, filename, "__init__.py")):
                rv.append(filename)

        for n, _ in list_ext(self.__plugin):
            rv.append(n)
        rv.sort()
        return rv

    def get_command(self, ctx, name):  # pylint: disable=arguments-differ
        command = super().get_command(ctx, name)
        if command is None:
            try:
                command = load_one(name, self.__plugin_folder, "cli", main=self)
            except FileNotFoundError:
                command = load_ext(name, self.__plugin, "cli", main=self)
        command.__name__ = name
        return command


async def node_attr(
    obj, path, attr, value=NotGiven, eval_=False, split_=False, res=None, chain=None
):
    """
    Sub-attr setter.

    Args:
        obj: command object
        path: address of the node to change
        attr: path of the element to change
        value: new value (default NotGiven)
        eval_: evaluate the new value? (default False)
        split_: split a string value into words? (bool or separator, default False)
        res: old node, if it has been read already
        chain: change chain of node, copied from res if clear

    Special: if eval_ is True, a value of NotGiven deletes, otherwise it
    prints the record without changing it. A mapping replaces instead of updating.

    Returns the result of setting the attribute, or ``None`` if it printed
    """
    if res is None:
        res = await obj.client.get(path, nchain=obj.meta or 2)
    if chain is None:
        chain = res.chain

    try:
        val = res.value
    except AttributeError:
        chain = None
    if split_ is True:
        split_ = ""
    if eval_:
        if value is NotGiven:
            value = res_delete(res, attr)
        else:
            value = eval(value)  # pylint: disable=eval-used
            if split_ is not False:
                value = value.split(split_)
            value = res_update(res, attr, value=value)
    else:
        if value is NotGiven:
            if not attr and obj.meta:
                val = res
            else:
                val = res_get(res, attr)
            yprint(val, stream=obj.stdout)
            return
        if split_ is not False:
            value = value.split(split_)
        value = res_update(res, attr, value=value)

    res = await obj.client.set(path, value=value, nchain=obj.meta, chain=chain)
    return res


def cmd():
    """
    The main command entry point, as declared in ``setup.py``.
    """
    try:
        # pylint: disable=no-value-for-parameter,unexpected-keyword-arg
        main(standalone_mode=False)
    except click.exceptions.MissingParameter as exc:
        print(f"You need to provide an argument { exc.param.name.upper() !r}.\n", file=sys.stderr)
        print(exc.cmd.get_help(exc.ctx), file=sys.stderr)
        sys.exit(2)
    except click.exceptions.UsageError as exc:
        try:
            s = str(exc)
        except TypeError:
            logger.exception(repr(exc), exc_info=exc)
        else:
            print(s, file=sys.stderr)
        sys.exit(2)
    except click.exceptions.Abort:
        print("Aborted.", file=sys.stderr)
        pass
    except EnvironmentError:  # pylint: disable=try-except-raise
        raise
    except (ClientError, ServerError) as err:
        print(type(err).__name__ + ":", *err.args, file=sys.stderr)
        sys.exit(1)


#   except BaseException as exc:
#       print(exc)
#       sys.exit(1)


@click.command(cls=partial(Loader, __file__, "command"))
@click.option(
    "-v", "--verbose", count=True, help="Enable debugging. Use twice for more verbosity."
)
@click.option(
    "-l", "--log", multiple=True, help="Adjust log level. Example: '--log asyncactor=DEBUG'."
)
@click.option("-q", "--quiet", count=True, help="Disable debugging. Opposite of '--verbose'.")
@click.option("-D", "--debug", is_flag=True, help="Enable debug speed-ups (smaller keys etc).")
@click.option("-c", "--cfg", type=click.File("r"), default=None, help="Configuration file (YAML).")
@click.option(
    "-C",
    "--conf",
    multiple=True,
    help="Override a config entry. Example: '-C server.bind_default.port=57586'",
)
@click.pass_context
async def main(ctx, verbose, quiet, debug, log, cfg, conf):
    """
    This is the DistKV command. You need to add a subcommand for it to do
    anything.

    You can print the current configuration with 'distkv dump cfg'.
    """
    ctx.ensure_object(attrdict)
    ctx.obj.debug = max(verbose - quiet + 1, 0)
    ctx.obj._DEBUG = debug
    ctx.obj.stdout = CFG.get("_stdout", sys.stdout)  # used for testing

    def _cfg(path):
        nonlocal cfg
        if cfg is not None:
            return
        if os.path.exists(path):
            try:
                cfg = open(path, "r")
            except PermissionError:
                pass

    _cfg(os.path.expanduser("~/config/distkv.cfg"))
    _cfg(os.path.expanduser("~/.config/distkv.cfg"))
    _cfg(os.path.expanduser("~/.distkv.cfg"))
    _cfg("/etc/distkv/distkv.cfg")
    _cfg("/etc/distkv.cfg")

    if cfg:
        logger.debug("Loading %s", cfg)

        cd = yload(cfg)
        if cd is None:
            ctx.obj.cfg = CFG
        else:
            ctx.obj.cfg = combine_dict(cd, CFG, cls=attrdict)
        cfg.close()
    else:
        ctx.obj.cfg = CFG

    # One-Shot-Hack the config file.
    for k in conf:
        try:
            k, v = k.split("=", 1)
        except ValueError:
            v = NotGiven
        else:
            try:
                v = eval(v)  # pylint: disable=eval-used
            except Exception:  # pylint: disable=broad-except
                pass
        c = ctx.obj.cfg
        *sl, s = k.split(".")
        for kk in sl:
            try:
                c = c[kk]
            except KeyError:
                c[kk] = attrdict()
                c = c[kk]
        if v is NotGiven:
            del c[s]
        else:
            c[s] = v

    # Configure logging. This is a somewhat arcane art.
    lcfg = ctx.obj.cfg.logging
    lcfg["root"]["level"] = (
        "DEBUG" if verbose > 2 else "INFO" if verbose > 1 else "WARNING" if verbose else "ERROR"
    )
    for k in log:
        k, v = k.split("=")
        lcfg["loggers"].setdefault(k, {})["level"] = v
    dictConfig(lcfg)
    logging.captureWarnings(verbose > 0)


@main.command(
    short_help="Import the debugger", help="Imports PDB and then continues to process arguments."
)
@click.argument("args", nargs=-1)
async def pdb(args):  # safe
    import pdb  # pylint: disable=redefined-outer-name

    pdb.set_trace()  # safe
    if not args:
        return
    return await main.main(args)
